import os

import pytpu as tpu


def get_data_path():
    return os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "../data"
    )


def get_programs_path(tpu_device: tpu.Device):
    dev_info = tpu_device.info()

    ddr_word_length = dev_info["axi_word_len"] // 8
    cache_word_length = dev_info["cache_word_len"] // 8
    programs_path = os.path.join(
        get_data_path(),
        "ddr_{}/cache_{}".format(ddr_word_length, cache_word_length)
    )
    if not os.path.isdir(programs_path):
        raise RuntimeError(
            "No device pytpu_tests for axi word length {} cache word length {}".format(ddr_word_length,
                                                                                       cache_word_length))
    return programs_path



import pytest

import pytpu as tpu


@pytest.mark.skip(reason="watchdog program has mem collision and breaks the device and other pytpu_tests")
def test_watchdog_error(tpu_device, watchdog_program):
    with pytest.raises(tpu.InferenceError):
        # infer failed program
        ...

    # @pytest.fixture(scope='module')
    # def watchdog_program(tpu_device: tpu.Device):
    #     program_path = "{}/watchdog.tpu".format(get_programs_path(tpu_device))
    #     with tpu_device.load(program_path) as tpu_program:
    #         yield tpu_program

    # # TODO: watchdog breaks the device?...
    # input_path = os.path.join(get_data_path(), "watchdog/input.bin")
    # inp = np.fromfile(input_path, dtype=np.int8)
    #
    # inference = TPUInference(watchdog_program)
    # inference.load([inp])
    # status = tpu_device.load_inference_sync(inference)
    # assert not status.is_success

# TODO: what about that?:
# @pytest.fixture(scope='module')
# def mem_collision_program(tpu_device: tpu.Device):
#     program_path = "{}/mem_collision.tpu".format(get_programs_path(tpu_device))
#     with tpu_device.load(program_path) as tpu_program:
#         yield tpu_program



# FIXME: refactor according to wiki?

import numpy as np
from PIL import Image


def get_img_sizes(input_data):
    sizes = []
    for inp in input_data:
        if len(inp) == 1:
            for image in inp:
                sizes.append(image.shape[:-1][::-1])
        else:
            sizes.append(inp[0].shape)
    return sizes


def preprocess_mnist(data, output_size):
    # channels = len(data[0].mode)
    channels = 1  # All Mnist networks requires 1 channel
    result = np.ndarray((0, *output_size, channels), dtype=np.float32)
    for img in data:
        img = img.resize(size=output_size, resample=Image.CUBIC)
        tensor = np.array(img).astype(np.float32)
        # tensor = cv2.resize(tensor, dsize=output_size, interpolation=cv2.INTER_CUBIC) # from openCV
        tensor = np.expand_dims(tensor, axis=0)
        if len(tensor.shape) < 4:
            tensor = np.expand_dims(tensor, axis=3)
        tensor = tensor / 255.0
        result = np.concatenate((result, tensor), 0)
    return result


def pre(path_to_image):
    image = Image.open(path_to_image)

    # run preprocessing data
    input_data = [(np.asarray(image).astype(np.float32),)]
    sizes = get_img_sizes(input_data)

    #
    images = []
    for data in input_data:
        for dat in data:
            images.append(Image.fromarray(np.uint8(dat)).convert('L'))

    _, dim0, dim1, _ = None, 28, 28, 1
    input_hw_size = dim0, dim1

    result_array = preprocess_mnist(images, input_hw_size)

    result_dict = {}
    input_tensor_names = ('input_input',)

    for input_tensor_name in input_tensor_names:
        result_dict[input_tensor_name] = result_array

    return result_dict, sizes


# Postprocessing
def get_slice_array_dict(idx_slice: int, inference_output):
    outs = list(inference_output.keys())
    out_dict = {}
    for i in range(len(outs)):
        out_dict[outs[i]] = inference_output[outs[i]][idx_slice:idx_slice + 1]

    return out_dict


def postprocess_imagenet(inference_results, class_offset: int):
    top = 1
    num_classes = []
    for result in inference_results:
        assert len(result.keys()) == 1
        input_data = result[list(result.keys())[0]]
        input_data = np.reshape(input_data, (input_data.shape[0], input_data.shape[-1]))
        classes = np.squeeze(input_data.argsort()[::, -top:])[()]
        # the operation '[()]' returns number if the array in 0-d and returns an array otherwise
        if not isinstance(classes, np.ndarray):
            classes = np.array([classes])
        classes = classes + class_offset

        num_classes.append(classes)

    return num_classes


def post(images_sizes, inference_output):
    predictions = []
    for idx in range(len(images_sizes)):
        outs_to_imagenet = get_slice_array_dict(idx, inference_output)

        prediction = postprocess_imagenet((outs_to_imagenet,), 0)
        predictions.append(prediction[0])

    return predictions



# FIXME: refactor according to wiki?

from typing import Tuple, Any

import numpy as np
from PIL import Image


def get_img_sizes(input_data):
    sizes = []
    for inp in input_data:
        if len(inp) == 1:
            for image in inp:
                sizes.append(image.shape[:-1][::-1])
        else:
            sizes.append(inp[0].shape)
    return sizes


# Preprocessing
def _central_crop(image: Image, percent: float = 1.0) -> Image:
    """Crop center from image."""
    height, width = np.shape(image)[:2]
    min_dim = int(min(height, width) * percent)
    crop_top = (height - min_dim) // 2
    crop_left = (width - min_dim) // 2
    return image.crop((crop_left, crop_top, crop_left + min_dim, crop_top + min_dim))


def _crop_and_resize(image: Image, target_img_size: Tuple[int, int]) -> Image:
    """Crop center from image and resize it to another shape."""
    new_img = _central_crop(image, 0.85)
    new_img = new_img.resize(size=target_img_size)
    return new_img


def preprocess_imagenet(data, output_size):
    channels = 3  # All Imagenet networks requires 3 channels
    result = np.ndarray((0, *output_size, channels), dtype=np.float32)
    for img in data:
        img = img.convert('RGB')
        img = _crop_and_resize(img, output_size)
        tensor = np.asarray(img).astype(np.float32)
        tensor = np.expand_dims(tensor, axis=0)
        result = np.concatenate((result, tensor), 0)
    return result


def _sub_mean_rgb(tensor):
    # RGB MEAN
    return tensor - np.array([123.68, 116.779, 103.939], dtype=np.float32)


def pre(image_path: Any):
    image = Image.open(image_path)

    # Data preprocessing
    input_data = [(np.asarray(image).astype(np.float32),)]
    sizes = get_img_sizes(input_data)

    images = []
    for data in input_data:
        for dat in data:
            images.append(Image.fromarray(np.uint8(dat)))

    _, dim0, dim1, _ = None, 224, 224, 3
    input_hw_size = dim0, dim1

    result_array = preprocess_imagenet(images, input_hw_size)
    result_array = _sub_mean_rgb(result_array)

    result_dict = {}

    input_tensor_names = ('Placeholder',)

    for input_tensor_name in input_tensor_names:
        result_dict[input_tensor_name] = result_array
    return result_dict, sizes


# Postprocessing
def get_slice_array_dict(idx_slice: int, inference_output):
    outs = list(inference_output.keys())
    out_dict = {}
    for i in range(len(outs)):
        out_dict[outs[i]] = inference_output[outs[i]][idx_slice:idx_slice + 1]
    return out_dict


def postprocess_imagenet(inference_results, class_offset, top):
    num_classes = []
    for result in inference_results:
        assert len(result.keys()) == 1
        input_data = result[list(result.keys())[0]]
        input_data = np.reshape(input_data, (input_data.shape[0], input_data.shape[-1]))
        classes = np.squeeze(input_data.argsort()[::, -top:])[()]
        if not isinstance(classes, np.ndarray):
            classes = np.array([classes])
        classes = classes + class_offset

        num_classes.append(classes)
    return num_classes


def post(images_sizes, inference_output):
    predictions = []
    for idx in range(len(images_sizes)):
        outs_to_imagenet = get_slice_array_dict(idx, inference_output)

        prediction = postprocess_imagenet((outs_to_imagenet,), 1, 1)
        predictions.append(prediction[0])

    return predictions


IMAGENET_CLASSES = {
    0: "background",
    1: "tench, Tinca tinca",
    2: "goldfish, Carassius auratus",
    256: "Leonberg",
    257: "Newfoundland, Newfoundland dog",
    258: "Great Pyrenees",
    817: "spindle",
    818: "sports car, sport car",
    819: "spotlight, spot",
    998: "bolete",
    999: "ear, spike, capitulum",
    1000: "toilet tissue, toilet paper, bathroom tissue",
}



import pytpu as tpu

from ._helpres import get_programs_path


def test_program_lut_parameters():
    tpu_devices = tpu.Device.list_devices()
    assert tpu_devices, 'No TPU device found, check /dev'
    with tpu.Device.open(tpu_devices[0]) as tpu_device:
        program_path = "{}/resnet50.tpu".format(get_programs_path(tpu_device))

        with tpu_device.load(program_path) as tpu_program:
            program_info = tpu_program.info()
            lut_activation_params = program_info['hardware_parameters']['lut_activation_params']
            assert lut_activation_params["input_bw"] == 12
            assert lut_activation_params["lut_bw"] == 16
            assert lut_activation_params["lut_depth"] == 8
            assert lut_activation_params["output_bw"] == 16



from ._bindings import Device
from ._bindings import Inference
from ._bindings import InferenceError
from ._bindings import Program
from ._converter_to_raw import convert_to_raw  # type: ignore

# FIXME: remove type ignore

__all__ = [
    'Device',
    'Inference',
    'Program',
    'InferenceError',
    'convert_to_raw',
]



from __future__ import annotations

__all__ = [
    'Device',
    'Program',
    'Inference',
    'InferenceError',
]

import collections.abc
import contextlib
import zipfile
import json
import os
import re
from ctypes import (CDLL, c_char, c_char_p, c_int, c_void_p, c_uint8, c_size_t, c_bool, cast,
                    POINTER, CFUNCTYPE, Structure)
from enum import Enum
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Mapping
from typing import Optional

import numpy as np

_LIB_TPU_SO_PATH = '/usr/lib/libtpu.2.so'


# libtpu = CDLL('libtpu-experimental/build/libtpu.so')
# libtpu = CDLL('/home/d.baburin/iva_tpu_sdk/libtpu.2/build/libtpu.2.so')


class _LibTPUBinder:
    """This class lazy binds to libtpu *.so file thus allowing module to be imported without *.so file"""

    def __init__(self):
        self._libtpu: Optional[CDLL] = None

    @property
    def libtpu(self) -> CDLL:
        if self._libtpu is None:
            self._bind()

        assert self._libtpu is not None

        return self._libtpu

    def _bind(self) -> None:
        libtpu = CDLL(_LIB_TPU_SO_PATH)

        _cb_pointer = CFUNCTYPE(None, POINTER(Inference), c_int, c_void_p)

        # Device
        libtpu.tpu_create_device.restype = POINTER(Device)
        # libtpu.tpu_create_device.argtypes = [c_int]
        libtpu.tpu_create_device.argtypes = [c_char_p]

        libtpu.tpu_destroy_device.restype = c_void_p
        libtpu.tpu_destroy_device.argtypes = [POINTER(Device)]

        libtpu.tpu_is_device_valid.restype = c_int
        libtpu.tpu_is_device_valid.argtypes = [POINTER(Device)]

        libtpu.tpu_init_device.restype = c_int
        libtpu.tpu_init_device.argtypes = [POINTER(Device)]

        libtpu.tpu_get_device_error_message.restype = c_char_p
        libtpu.tpu_get_device_error_message.argtypes = [POINTER(Device)]

        libtpu.tpu_get_device_info.restype = c_char_p
        libtpu.tpu_get_device_info.argtypes = [POINTER(Device)]

        libtpu.tpu_load_program.restype = c_int
        libtpu.tpu_load_program.argtypes = [POINTER(Device), POINTER(Program)]

        libtpu.tpu_run_inference.restype = c_int
        libtpu.tpu_run_inference.argtypes = [POINTER(Device), POINTER(Inference), c_int]

        libtpu.tpu_submit_inference.restype = c_int
        libtpu.tpu_submit_inference.argtypes = [POINTER(Device), POINTER(Inference), c_int, _cb_pointer, c_void_p]

        # Program
        libtpu.tpu_create_program.restype = POINTER(Program)
        libtpu.tpu_create_program.argtypes = [POINTER(c_char), c_int]

        libtpu.tpu_destroy_program.restype = c_void_p
        libtpu.tpu_destroy_program.argtypes = [POINTER(Program)]

        libtpu.tpu_is_program_valid.restype = c_int
        libtpu.tpu_is_program_valid.argtypes = [POINTER(Program)]

        libtpu.tpu_get_program_info.restype = c_char_p
        libtpu.tpu_get_program_info.argtypes = [POINTER(Program)]

        libtpu.tpu_get_program_error_message.restype = c_char_p
        libtpu.tpu_get_program_error_message.argtypes = [POINTER(Program)]

        libtpu.tpu_get_batch_size.restype = c_size_t
        libtpu.tpu_get_batch_size.argtypes = [POINTER(Program)]

        libtpu.tpu_get_input_count.restype = c_size_t
        libtpu.tpu_get_input_count.argtypes = [POINTER(Program)]

        libtpu.tpu_get_input_size.restype = c_size_t
        libtpu.tpu_get_input_size.argtypes = [POINTER(Program), c_size_t, c_bool]

        libtpu.tpu_get_output_count.restype = c_size_t
        libtpu.tpu_get_output_count.argtypes = [POINTER(Program)]

        libtpu.tpu_get_output_size.restype = c_size_t
        libtpu.tpu_get_output_size.argtypes = [POINTER(Program), c_size_t, c_bool]

        # _TensorBufferObject
        libtpu.tpu_create_tensor_buffer_object.restype = POINTER(_TensorBufferObject)
        libtpu.tpu_create_tensor_buffer_object.argtypes = [POINTER(Program), c_int]

        libtpu.tpu_destroy_tensor_buffer_object.restype = c_void_p
        libtpu.tpu_destroy_tensor_buffer_object.argtypes = [POINTER(_TensorBufferObject)]

        libtpu.tpu_process_tensor_buffers.restype = c_void_p
        libtpu.tpu_process_tensor_buffers.argtypes = [POINTER(_TensorBufferObject)]

        libtpu.tpu_get_tensor_buffer_ptr.restype = POINTER(c_uint8)
        libtpu.tpu_get_tensor_buffer_ptr.argtypes = [POINTER(_TensorBufferObject), c_size_t, c_bool]

        libtpu.tpu_set_user_tensor_buffer_ptr.restype = POINTER(c_uint8)
        libtpu.tpu_set_user_tensor_buffer_ptr.argtypes = [POINTER(_TensorBufferObject), c_size_t, POINTER(c_uint8)]

        # Inference
        libtpu.tpu_create_inference.restype = POINTER(Inference)
        libtpu.tpu_create_inference.argtypes = [POINTER(Program)]

        libtpu.tpu_destroy_inference.restype = c_void_p
        libtpu.tpu_destroy_inference.argtypes = [POINTER(Inference)]

        libtpu.tpu_get_inference_program.restype = POINTER(Program)
        libtpu.tpu_get_inference_program.argtypes = [POINTER(Inference)]

        libtpu.tpu_get_inference_inputs.restype = POINTER(_TensorBufferObject)
        libtpu.tpu_get_inference_inputs.argtypes = [POINTER(Inference)]

        libtpu.tpu_set_inference_inputs.restype = POINTER(_TensorBufferObject)
        libtpu.tpu_set_inference_inputs.argtypes = [POINTER(Inference), POINTER(_TensorBufferObject)]

        libtpu.tpu_get_inference_outputs.restype = POINTER(_TensorBufferObject)
        libtpu.tpu_get_inference_outputs.argtypes = [POINTER(Inference)]

        libtpu.tpu_set_inference_outputs.restype = POINTER(_TensorBufferObject)
        libtpu.tpu_set_inference_outputs.argtypes = [POINTER(Inference), POINTER(_TensorBufferObject)]

        libtpu.tpu_get_inference_status.restype = c_int
        libtpu.tpu_get_inference_status.argtypes = [POINTER(Inference)]

        libtpu.tpu_get_inference_error_message.restype = c_char_p
        libtpu.tpu_get_inference_error_message.argtypes = [POINTER(Inference)]

        self._libtpu = libtpu


_LIB_TPU_BINDER = _LibTPUBinder()


class ProcessingMode(Enum):
    RAW = 0
    PRE_PROCESS = 1
    POST_PROCESS = 2
    FULL = 3


class _TensorBufferType(Enum):
    INPUTS = 0
    OUTPUTS = 1


_Pointer = Any


class Device(Structure):
    _fields_ = []

    @staticmethod
    @contextlib.contextmanager
    def open(dev_name: str) -> Generator[Device, None, None]:
        """Open TPU device with a give name. This is context manager function (generator) typically used in `with`
         statement:

            with Device.open('/dev/tpu0) as device: ...

        """
        if not isinstance(dev_name, str):
            raise ValueError(f'Device name must be a string: {type(dev_name)}')

        pointer = _LIB_TPU_BINDER.libtpu.tpu_create_device(c_char_p(dev_name.encode('utf-8')))
        device = Device(pointer)

        yield device
        _LIB_TPU_BINDER.libtpu.tpu_destroy_device(pointer)

    @staticmethod
    def list_devices() -> List[str]:
        # TODO: check why len(f) == 4?
        return [os.path.join('/dev', f) for f in os.listdir('/dev') if re.match(r'tpu.', f) and len(f) == 4]

    def __init__(self, pointer: _Pointer) -> None:
        """TPU device descriptor that holds information about to the currently opened TPU device (like a descriptor of
         an opened file). This method should not be directly called by the user, instead, use `Device.open` method."""

        super().__init__()

        # TODO: check if pointer is actually a string and remind user to use Device.open method instead

        self._pointer = pointer

        init_code = _LIB_TPU_BINDER.libtpu.tpu_init_device(pointer)
        if init_code != 0:
            raise ValueError(f'Invalid init code {init_code}: {self._get_error_message()}')

        if not _LIB_TPU_BINDER.libtpu.tpu_is_device_valid(self._pointer):
            raise ValueError(f'Invalid device: {self._get_error_message()}')

    def _get_error_message(self) -> str:
        msg = _LIB_TPU_BINDER.libtpu.tpu_get_device_error_message(self._pointer)
        return msg.decode('utf-8')

    # TODO: user https://mypy.readthedocs.io/en/latest/more_types.html#typeddict ?
    def info(self) -> Dict:
        return json.loads(_LIB_TPU_BINDER.libtpu.tpu_get_device_info(self._pointer).decode('utf-8'))

    def close(self) -> None:
        _LIB_TPU_BINDER.libtpu.tpu_destroy_device(self._pointer)

    @contextlib.contextmanager
    def load(self, program_path: str) -> Generator[Program, None, None]:
        """Load program from a given path to .tpu file into TPU. This allocates a certain portion of
         devices RAM memory. This is context manger (generator) and upon exiting the program will be
          unloaded from TPU and memory released."""
        is_raw = _is_raw_program(program_path)

        if not isinstance(program_path, str):
            raise ValueError(f'Program path must be a string: {type(program_path)}')

        program_pointer = _LIB_TPU_BINDER.libtpu.tpu_create_program(c_char_p(program_path.encode('utf-8')), is_raw)
        program = Program(self, program_pointer, is_raw)
        _LIB_TPU_BINDER.libtpu.tpu_load_program(self._pointer, program_pointer)

        yield program
        _LIB_TPU_BINDER.libtpu.tpu_destroy_program(program_pointer)


class _TensorBufferObject(Structure):
    _fields_ = []

    def __init__(self, pointer: _Pointer) -> None:
        super().__init__()
        self._pointer = pointer

    def set_user_tensor_buffer_ptr(self, n: int, ptr: _Pointer) -> _Pointer:
        return _LIB_TPU_BINDER.libtpu.tpu_set_user_tensor_buffer_ptr(self._pointer, c_size_t(n), ptr)

    # def get_tensor_buffer_ptr(self, n: int, is_raw: bool) -> _Pointer:
    #     return _LIB_TPU_BINDER.libtpu.tpu_get_tensor_buffer_ptr(self._pointer, c_size_t(n), c_bool(is_raw))


class Program(Structure):
    _fields_ = []

    def __init__(self, device: Device, pointer: _Pointer, is_raw: bool = False) -> None:
        """Descriptor of the currently loaded program. This method should not be called directly by the user, instead,
         load a program into a give instance of TPU device with `Device.load` method."""
        super().__init__()

        self._device = device
        self._pointer = pointer
        self._is_raw = is_raw

        # Check status
        if not _LIB_TPU_BINDER.libtpu.tpu_is_program_valid(self._pointer):
            msg = _LIB_TPU_BINDER.libtpu.tpu_get_program_error_message(self._pointer)
            raise ValueError('Invalid program: ' + str(msg.decode('utf-8')))

        # Preload metadata
        metadata = self.info()
        self._input_metadata = metadata['inputs']['1']
        self._output_metadata = metadata['outputs']['2']
        self._hardware_parameters_metadata = metadata['hardware_parameters']

    # TODO: Use https://mypy.readthedocs.io/en/latest/more_types.html#typeddict here?
    def info(self) -> Dict:
        return json.loads(_LIB_TPU_BINDER.libtpu.tpu_get_program_info(self._pointer).decode('utf-8'))

    def get_input_size(self, n: int, is_raw: bool) -> int:
        return _LIB_TPU_BINDER.libtpu.tpu_get_input_size(self._pointer, c_size_t(n), c_bool(is_raw))

    def get_output_size(self, n: int, is_raw: bool) -> int:
        return _LIB_TPU_BINDER.libtpu.tpu_get_output_size(self._pointer, c_size_t(n), c_bool(is_raw))

    def close(self) -> None:
        _LIB_TPU_BINDER.libtpu.tpu_destroy_program(self._pointer)

    @contextlib.contextmanager
    def inference(self) -> Generator[Inference, None, None]:
        """Initialize inference session. Yields inference objects that manages multiple sequential inferences.
         The instance of Inference class is meant to be reused multiple times. """
        inference = Inference(self)
        yield inference
        inference.close()

    @contextlib.contextmanager
    def buffer(self, type_: _TensorBufferType) -> Generator[_TensorBufferObject, None, None]:
        buffer_pointer = _LIB_TPU_BINDER.libtpu.tpu_create_tensor_buffer_object(self._pointer,
                                                                                type_.value)  # make pointer
        yield _TensorBufferObject(buffer_pointer)

        # TODO: Calling tpu_destroy_tensor_buffer_object makes some tests unstable,
        #  see https://jira.hi-tech.org/browse/TPU2-2010
        _LIB_TPU_BINDER.libtpu.tpu_destroy_tensor_buffer_object(buffer_pointer)  # destroy pointer


CTX = Any


class InferenceError(ValueError):
    """Exception raised if libtpu reports inference as invalid"""


class Inference(Structure):

    _fields_ = []

    def __init__(self, program: Program):
        """Descriptor of inference session. This method should not be called directly, instead, initialize inference
         session with a give TPU program with `Program.inference`."""
        super().__init__()
        self._program = program
        self._pointer = _LIB_TPU_BINDER.libtpu.tpu_create_inference(program._pointer)

        # Explicitly order inputs and outputs:
        _input_tensor_name_to_input_tensor_index = {
            tensor_name: i
            for i, tensor_name in enumerate(program._input_metadata)
        }

        _output_tensor_name_to_output_tensor_index = {
            tensor_name: i
            for i, tensor_name in enumerate(program._output_metadata)
        }

        self._tensor_name_to_tensor_index: Dict[str, int] = dict()
        self._tensor_name_to_tensor_index.update(_input_tensor_name_to_input_tensor_index)
        self._tensor_name_to_tensor_index.update(_output_tensor_name_to_output_tensor_index)

        # Precalculate sizes:
        _input_sizes = {
            tensor_name: program.get_input_size(self._tensor_name_to_tensor_index[tensor_name], self._program._is_raw)
            for tensor_name in program._input_metadata
        }

        _output_sizes = {
            tensor_name: program.get_output_size(self._tensor_name_to_tensor_index[tensor_name], self._program._is_raw)
            for tensor_name in program._output_metadata
        }

        self._tensor_name_to_tensor_size: Dict[str, int] = dict()
        self._tensor_name_to_tensor_size.update(_input_sizes)
        self._tensor_name_to_tensor_size.update(_output_sizes)

    def _run(self, mode: ProcessingMode) -> None:
        status_code = _LIB_TPU_BINDER.libtpu.tpu_run_inference(
            self._program._device._pointer,  # device pointer
            self._pointer,  # inference pointer
            mode.value,  # pass numerical value of processing mode
        )

        if status_code != 0:
            msg = self._get_error_message()
            raise InferenceError(f'Invalid tpu_run_inference return code {status_code}: {msg}')

    def _get_error_message(self) -> str:
        return _LIB_TPU_BINDER.libtpu.tpu_get_inference_error_message(self._pointer).decode('utf-8')

    def sync(self, inputs, mode: ProcessingMode = ProcessingMode.FULL) -> Dict[str, np.ndarray]:

        # TODO: check that input is indeed a dictionary and each value has a valid dtype!

        with contextlib.ExitStack() as buffer_exit_stack:
            # Reserve pointers for input and output buffers:
            tpu_input_buf = buffer_exit_stack.enter_context(self._program.buffer(_TensorBufferType.INPUTS))
            tpu_output_buf = buffer_exit_stack.enter_context(self._program.buffer(_TensorBufferType.OUTPUTS))

            # Load
            ctx = self._load_inference(tpu_input_buf, tpu_output_buf, inputs)  # noqa
            # need to keep here ctx to prevent
            # being deleted by the garbage collector

            # Run
            self._run(mode)

            # # Get results:
            # result = self._get_inference(tpu_output_buf)
            result = self._get_inference(ctx)

            # NOTE: only after we have a copy of results in memory as numpy arrays -
            #       we can exit from buffer_exit_stack thus releasing IO pointers!

            with contextlib.redirect_stdout(None):
                print(ctx)  # TODO: consider redesign; need to keep ctx from being collected by the GC

        self._check_status()

        return result

    def _check_status(self) -> None:
        status_code = _LIB_TPU_BINDER.libtpu.tpu_get_inference_status(self._pointer)
        if status_code != 0:
            msg = self._get_error_message()
            raise InferenceError(f'Invalid inference status code {status_code}: {msg}')

    def close(self):
        _LIB_TPU_BINDER.libtpu.tpu_destroy_inference(self._pointer)

    def _load_inference(self, tpu_input: _TensorBufferObject, tpu_output: _TensorBufferObject,
                        input_data: Mapping[str, np.ndarray]) -> Dict[str, CTX]:

        if not isinstance(input_data, collections.abc.Mapping):
            raise ValueError('Input data must be a mapping, got:' + str(type(input_data)))

        # Inputs:
        for tensor_name, tensor_data in input_data.items():
            tensor_idx = self._tensor_name_to_tensor_index[tensor_name]
            tensor_size = self._tensor_name_to_tensor_size[tensor_name]

            tpu_input.set_user_tensor_buffer_ptr(tensor_idx, tensor_data.ctypes.data_as(POINTER(c_uint8)))
            _LIB_TPU_BINDER.libtpu.tpu_set_inference_inputs(self._pointer, tpu_input._pointer)

        # Outputs:
        out_ctx = dict()

        for tensor_name in self._program._output_metadata:
            tensor_idx = self._tensor_name_to_tensor_index[tensor_name]
            tensor_size = self._tensor_name_to_tensor_size[tensor_name]

            out_tensor = bytearray(np.empty((tensor_size,), dtype=np.uint8))
            p_out_tensor = (c_uint8 * tensor_size).from_buffer(out_tensor)
            c_out_ptr = cast(p_out_tensor, POINTER(c_uint8))
            tpu_output.set_user_tensor_buffer_ptr(tensor_idx, c_out_ptr)

            out_ctx[tensor_name] = c_out_ptr

        _LIB_TPU_BINDER.libtpu.tpu_set_inference_outputs(self._pointer, tpu_output._pointer)

        return out_ctx

    # def _get_inference(self, tpu_output_buf: _TensorBufferObject) -> Dict[str, np.ndarray]:
    def _get_inference(self, ctx: Dict[str, CTX]) -> Dict[str, np.ndarray]:
        out_meta = self._program._output_metadata
        hw_par_meta = self._program._hardware_parameters_metadata

        collected_data = dict()
        # is_raw = self._program._is_raw

        for tensor_name, out_ctx in ctx.items():
            tensor_size = self._tensor_name_to_tensor_size[tensor_name]
            data = np.ctypeslib.as_array(out_ctx, shape=(tensor_size,))
        # for tensor_name in out_meta:
            # tensor_idx = self._tensor_name_to_tensor_index[tensor_name]
            # _ctx = tpu_output_buf.get_tensor_buffer_ptr(tensor_idx, is_raw)
            # data = np.ctypeslib.as_array(_ctx, shape=(tensor_size,))

            # if is_raw:
            if self._program._is_raw:
                data = data.reshape(-1, hw_par_meta["ddr_word_len"])
            else:
                data = _convert_from_uint8_to_user_dtype(data, out_meta[tensor_name]['user_dtype'])
                data = data.reshape(out_meta[tensor_name]['user_shape'])

            collected_data[tensor_name] = data

        return collected_data


def _convert_from_uint8_to_user_dtype(data, user_dtype):
    if 'int' in user_dtype:
        return data.view(user_dtype)
    elif 'float16' == user_dtype:
        return data.view(np.uint16).view(user_dtype)
    elif 'float32' == user_dtype:
        return data.view(np.uint32).view(user_dtype)
    else:
        raise AssertionError(f'{user_dtype} as user_dtype does not support!')


def _is_raw_metadata(metadata: Dict[str, Any]) -> bool:
    """Heuristic to determine if the metadata dictionary corresponds to the RAW program."""

    # Metadata for outputs (first argv region in the outputs)
    out_meta = next(iter(metadata['outputs'].values()))  # type: ignore

    # Metadata for the first output tensor (no mixed raw/non-raw tensor metadata is possible so any output tensor metadata will suffice):
    output_tensor_metadata = next(iter(out_meta.values()))

    # RAW programs only contain 'address' and 'size' fields, user_* fields are only present in non-raw programs:
    return 'user_dtype' not in output_tensor_metadata


def _is_raw_program(program_path: str) -> bool:
    """Inspect ZIP file (open with zipfile) and determine determine whether program is in RAW format or not."""

    with zipfile.ZipFile(program_path, 'r') as tpu_program:
        data = tpu_program.read('metadata.json')

    metadata = json.loads(data)
    return _is_raw_metadata(metadata)



import concurrent.futures
import contextlib
import os.path
import threading
from collections import deque
from itertools import cycle
from math import ceil

import pytest

import pytpu as tpu
from ._helpres import get_data_path, get_programs_path


from ._resnet50 import pre as resnet50_pre
from ._resnet50 import post as resnet50_post

testdata = [
    (os.path.join(get_data_path(), "resnet50/images/0.jpg"), 227),
]


@pytest.mark.parametrize("image_path, expected", testdata)
def test_resnet50_inference_sync(image_path: str, expected: int):
    num_images = 100

    def pre(image_path):
        tensor_dct, sizes = resnet50_pre(image_path)
        return tensor_dct, sizes

    def post(out, sizes):
        return resnet50_post(sizes, out)

    inputs = [image_path for _ in range(num_images)]

    with contextlib.ExitStack() as tpu_device_stack:
        tpu_devices = [tpu_device_stack.enter_context(tpu.Device.open(i)) for i in tpu.Device.list_devices()]
        program_path = "{}/resnet50.tpu".format(get_programs_path(tpu_devices[0]))

        with contextlib.ExitStack() as tpu_program_stack:
            tpu_programs = [tpu_program_stack.enter_context(tpu.load(program_path)) for tpu in tpu_devices]
            with contextlib.ExitStack() as tpu_inference_stack:
                programs_merged_inputs = zip(cycle(tpu_programs), range(len(inputs)))

                for tpu_program, input_index in programs_merged_inputs:
                    inference = tpu_inference_stack.enter_context(tpu_program.inference())
                    input_, sizes = pre(inputs[input_index])
                    out = inference.sync(input_)

                    result = post(out, sizes)[0]
                    assert result == expected


@pytest.mark.skip()  # TODO: add appropriate JIRA issue?
@pytest.mark.parametrize("image_path, expected", testdata)
def test_resnet50_inference_thread(image_path: str, expected: int):
    _thread_to_inference = dict()

    max_workers = 4
    num_images = 100

    def pre(image_path):
        tensor_dct, sizes = resnet50_pre(image_path)
        return tensor_dct, sizes

    def post(out, sizes):
        return resnet50_post(sizes, out)

    def worker_init():
        tpu_program = tpu_program_queque.pop()
        thread = threading.current_thread()
        _thread_to_inference[thread] = tpu.Inference(tpu_program)

    def target(path):
        input_, sizes = pre(path)
        thread = threading.current_thread()
        inference = _thread_to_inference[thread]
        out = inference.sync(input_)
        result = post(out, sizes)
        return result

    with contextlib.ExitStack() as tpu_device_stack:
        tpu_devices = [tpu_device_stack.enter_context(tpu.Device.open(i)) for i in tpu.Device.list_devices()]

        with contextlib.ExitStack() as tpu_program_stack:
            program_path = "{}/resnet50.tpu".format(get_programs_path(tpu_devices[0]))
            tpu_programs = [tpu_program_stack.enter_context(tpu.load(program_path)) for tpu in tpu_devices]

            tpu_program_queque = deque(tuple(tpu_programs) * ceil(max_workers / len(tpu_devices)))
            inputs = [image_path for _ in range(num_images)]
            try:
                with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers, initializer=worker_init,
                                                           initargs=()) as executor:
                    futures = []

                    for path in inputs:
                        futures.append(executor.submit(target, path))

                    for future in concurrent.futures.as_completed(futures):
                        out = future.result()
                        assert out[0] == expected

            finally:
                for inference in _thread_to_inference.values():
                    inference.close()



import pytpu as tpu


def test_total_mem():
    tpu_devices = tpu.Device.list_devices()
    assert tpu_devices, 'No TPU device found, check /dev'
    with tpu.Device.open(tpu_devices[0]) as tpu_device:
        device_info = tpu_device.info()
        assert device_info['cache_buf_depth'] > 0
        assert device_info['cache_buf_number'] > 0
        assert device_info['ewp_mem_n_buf'] > 0
        assert device_info['ewp_mem_buf_depth'] > 0



import warnings

from setuptools import find_packages
from setuptools import setup

VERSION = "15.0.14"  # WARNING: build script replace 15.0.14 with an actual three-decimal version using `sed` in file!

if not VERSION or VERSION == "15.0.14":
    warnings.warn(f'Invalid package version "{VERSION}". If that is NOT a local debug installation you must replace '
                  f'VERSION with a valid value or provide 15.0.14 environment variable for the build script.')

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='pytpu',
    packages=find_packages(exclude='pytpu_tests'),
    version=VERSION,
    author="IVA-Tech",
    author_email="info@iva-tech.ru",
    description="TPU Python API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://git.mmp.iva-tech.ru/tpu_sw/iva_tpu_sdk",
    install_requires=[
        'numpy',
    ],
    extras_require={
        'test': [
            'pytest',
            'pytest-xdist',
            'pytest-cov',
            'Pillow==9.5.0',
            'flake8',
            'mypy',

            # Documentation
            'sphinx',  # docs engine
            'sphinx-click',  # to automatically document click-based applications
            'sphinx-rtd-theme',  # ReadTheDocs theme for html formatting
            'sphinxcontrib-plantuml',  # UML diagrams support

            'tpu-tlm-is~=0.3.0.1'  # To test RAW mode
        ],
    },
    zip_safe=False,
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'run_get_fps = pytpu.scripts.run_get_fps:main',
            'pyrun_tpu = pytpu.scripts.pyrun_tpu_cli:main'
        ]
    },
)



# type: ignore

from __future__ import annotations

# TODO: fix type ignore

__all__ = [
    'convert_to_raw'
]

import copy
import json
import os
import shutil
import tempfile
import zipfile
from math import ceil
from typing import Any
from typing import Dict
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from tpu_tlm_is.base import TensorDescription
    from tpu_tlm_is.base import TpuParameters


class Codec:
    def __init__(self, tpu_par: TpuParameters, tensor_descriptions: Dict[str, TensorDescription]) -> None:
        self._tpu_par = tpu_par
        self._tensor_descriptions = tensor_descriptions

    def encode(self, input_dict: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        from tpu_tlm_is.models.iotools import tpu_encode
        return tpu_encode(self._tpu_par, input_dict, self._tensor_descriptions)

    def decode(self, output_raw_dict: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        from tpu_tlm_is.models.iotools import tpu_decode
        return tpu_decode(self._tpu_par, output_raw_dict, self._tensor_descriptions)


def _to_raw_old(io_: Dict[str, Any]) -> Dict[str, Any]:
    return {
        'address': io_['address'],
        'size': io_['size'],
    }


def _get_tpu_parameters(hw_par: Dict[str, Any]) -> TpuParameters:
    from tpu_tlm_is.base import get_hw_params

    if hw_par['cache_word_len'] == 128 and hw_par['ddr_word_len'] == 32:
        return get_hw_params('128x128').tpu_parameters
    if hw_par['cache_word_len'] == 64 and hw_par['ddr_word_len'] == 16:
        return get_hw_params('64x64').tpu_parameters
    if hw_par['cache_word_len'] == 64 and hw_par['ddr_word_len'] == 64:
        return get_hw_params('64x64_fpga').tpu_parameters

    raise NotImplementedError('Translation of such platform does not supported!')


def _to_raw(metadata: Dict[str, Any]) -> Dict[str, Any]:
    raw_metadata = copy.copy(metadata)
    for idx, region in raw_metadata['inputs'].items():
        for name, _ in region.items():
            raw_metadata['inputs'][idx][name] = _to_raw_old(metadata['inputs'][idx][name])

    for idx, region in raw_metadata['outputs'].items():
        for name, _ in region.items():
            raw_metadata['outputs'][idx][name] = _to_raw_old(metadata['outputs'][idx][name])

    return raw_metadata


def _get_tensor_description(io_: Dict[str, Any], cwl: int) -> TensorDescription:
    from tpu_tlm_is.base import TensorDescription
    from tpu_tlm_is.base.number_types import STR_TO_USER_NUMBER_TYPE
    from tpu_tlm_is.base.number_types import STR_TO_TPU_NUMBER_TYPE
    from tpu_tlm_is.base.number_types import UserNumberType
    from tpu_tlm_is.base.number_types import TpuNumberType

    if 'user_shape' in io_.keys():
        return TensorDescription(
            user_shape_mask=tuple(tuple([True, ] * abs(p[0]) + [False, ] * s + [True, ] * abs(p[1])
                                        for p, s in zip(io_['padding'], io_['user_shape']))),
            user_order=io_['user_order'],
            user_dtype=STR_TO_USER_NUMBER_TYPE[io_['user_dtype']],
            tpu_shape=io_['tpu_shape'],
            tpu_order=io_['tpu_order'],
            tpu_dtype=STR_TO_TPU_NUMBER_TYPE[io_['tpu_dtype']],
            scales=tuple([float(s) for s in io_['scales']]),
            anchor=io_['anchor'],
        )
    else:
        return TensorDescription(
            user_shape_mask=((False,), tuple([False, ] * int(io_['size'])),),
            user_order=('N', 'C',),
            user_dtype=UserNumberType.INT8,
            tpu_shape=(1, ceil(int(io_['size']) / cwl), np.minimum(cwl, int(io_['size']))),
            tpu_order=('N', 'C', 'B'),
            tpu_dtype=TpuNumberType.INT8,
            scales=(1.0,),
        )


def convert_to_raw(program_path: str, raw_program_path: str) -> Codec:
    try:
        from tpu_tlm_is.base import TensorDescription
    except ImportError:
        raise ImportError('Raw conversion is a debug feature that requires internal'
                          ' library tpu_tlm_is to be installed.')

    with zipfile.ZipFile(program_path, 'r') as program:
        data = program.read('metadata.json')

    metadata = json.loads(data)
    tpu_par = _get_tpu_parameters(metadata['hardware_parameters'])
    tensor_descriptions: Dict[str, TensorDescription] = dict()

    for _, region in metadata['inputs'].items():
        for name, io_ in region.items():
            tensor_descriptions[name] = _get_tensor_description(io_, tpu_par.cache_word_length)

    for _, region in metadata['outputs'].items():
        for name, io_ in region.items():
            tensor_descriptions[name] = _get_tensor_description(io_, tpu_par.cache_word_length)

    # TODO: rewrite later without extracting all zip file to a temp directory

    with tempfile.TemporaryDirectory() as tempdir:
        with zipfile.ZipFile(program_path, 'r') as zip_obj:
            zip_obj.extractall(tempdir)

        with open(os.path.join(tempdir, 'metadata.json'), 'w') as metadata_file:
            json.dump(_to_raw(metadata), metadata_file, indent=4)

        shutil.make_archive(raw_program_path, 'zip', tempdir)
        os.rename(raw_program_path + '.zip', raw_program_path)

    return Codec(tpu_par, tensor_descriptions)
